/*
 * POP3 Client Program
 *
 * cc -o biff biff.c
 *
 * Rev1.0 2014/11/15 First release
 * Rev1.1 2014/11/19 add net_list()
 * Rev1.2 2015/04/20 change net_list()
 *
 */
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <errno.h>
#include <netdb.h>
#include <ctype.h>
#include <syslog.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <openssl/ssl.h>
#include <openssl/err.h>


#define BUFMAX  256
#define	CMDMAX	128
#define	CNFMAX	64
#define POP	110
#define POPS	995
#define DEBUG	0

#define MIN(a, b) ((a) < (b) ? (a) : (b))
#define MAX(a, b) ((a) > (b) ? (a) : (b))

extern int h_errno;
int net;
SSL *ssl=NULL;
SSL_CTX *ctx=NULL;

int nbyte;
char pop_cmd[CMDMAX];
char r_msg[BUFMAX];
char buf[BUFMAX];
char wkbf[81];

typedef struct {
    char from[CNFMAX];
    char title[CNFMAX];
    int del;
    int exit;
} mail;

mail mailwk;

typedef struct {
    int port;
    char cdir[CNFMAX];
    char nkf[CNFMAX];
    char host[CNFMAX];
    char user[CNFMAX];
    char pass[CNFMAX];
    int mailn;
    mail *mailp;
} configuration;

configuration config;

void sys_error(char *func){
    openlog("biff", LOG_CONS | LOG_PID, LOG_USER);
    syslog(LOG_NOTICE, func);
    closelog();
}

void net_error(char *func){
//    openlog("biff", LOG_CONS | LOG_PID, LOG_USER);
//    syslog(LOG_NOTICE, func);
//    closelog();
    sys_error(func);

    if (ssl != NULL) {
      SSL_shutdown(ssl);
      SSL_free(ssl);		//SSL_new()で確保した領域解放
      SSL_CTX_free(ctx);	//SSL_CTX_new()で確保した領域解放
      ERR_free_strings();	//SSL_load_error_strings()で確保した領域解放
    }
    shutdown(net, 2);
    close(net);
    exit(0);
}

int ini_parse(char * cdir, char * file) {
    FILE *fp;
    char wk[3];
    int msize;

    if ((fp = fopen(file, "r")) == NULL) {
      printf("file open error!!\n");
      return(1);
    }

    config.port=POP;
    strcpy(config.cdir,cdir);
    memset(&mailwk,0,sizeof(mailwk));
    while (fgets(wkbf, 256, fp) != NULL) {
      if (wkbf[0] == '#') continue;
if(DEBUG)printf("%s", wkbf);
      if(strncmp(wkbf,"SSL=ON",6) == 0) config.port=POPS;
      if(strncmp(wkbf,"HOST=",5) == 0) 
        strncpy(config.host,&wkbf[5],MIN(strlen(wkbf)-6,CNFMAX-1));
      if(strncmp(wkbf,"USER=",5) == 0) 
        strncpy(config.user,&wkbf[5],MIN(strlen(wkbf)-6,CNFMAX-1));
      if(strncmp(wkbf,"PASS=",5) == 0) 
        strncpy(config.pass,&wkbf[5],MIN(strlen(wkbf)-6,CNFMAX-1));
      if(strncmp(wkbf,"TITLE=",6) == 0) 
        strncpy(mailwk.title,&wkbf[6],MIN(strlen(wkbf)-7,CNFMAX-1));
      if(strncmp(wkbf,"FROM=",5) == 0) 
        strncpy(mailwk.from,&wkbf[5],MIN(strlen(wkbf)-6,CNFMAX-1));
      if(strncmp(wkbf,"DEL=ON",6) == 0) mailwk.del=1;
      if(strncmp(wkbf,"EXIT=",5) == 0) {
        memset(wk,0,sizeof(wk));
        wk[0]=wkbf[5];
        if (wkbf[6] >= '0' && wkbf[6] <= '9') wk[1]=wkbf[6];
        mailwk.exit=atoi(wk);
        msize=(config.mailn+1)*sizeof(mail);
if(DEBUG)printf("config.mailp=%p config.mailn=%d msize=%d sizeof(mail)=%d\n",config.mailp,config.mailn,msize,sizeof(mail));
//        config.mailp=(struct mail *)realloc(config.mailp,msize);
        config.mailp=realloc(config.mailp,msize);
if(DEBUG)printf("config.mailp=%p\n",config.mailp);
        memcpy(&config.mailp[config.mailn],&mailwk,sizeof(mailwk));
        config.mailn++;
        memset(&mailwk,0,sizeof(mailwk));
      }
    }
    fclose(fp);
    return(0);
}

int nkfPath(char * obuf, int olen){
    FILE *fp;
    char cmd[20];

    strcpy(cmd,"whereis nkf");
if(DEBUG)printf("nkfPath:cmd=[%s]\n",cmd);
    if ( (fp=popen(cmd,"r")) == NULL) {
      sys_error("nkfPath:popen()");
      return(0);
    }

    while(fgets(wkbf, sizeof(wkbf), fp) == NULL) {
      sys_error("nkfPath:fgets()");
      return(0);
    }

    pclose(fp);
    if (strncmp(wkbf,"nkf:",4) != 0) return(0);
if(DEBUG)printf("nkfPath:wkbf(1)=[%s] strlen=%d\n",wkbf,strlen(wkbf));
    wkbf[strlen(wkbf)-1]=0;
if(DEBUG)printf("nkfPath:wkbf(2)=[%s] strlen=%d\n",wkbf,strlen(wkbf));
    strncpy(obuf,&wkbf[5],MIN(strlen(&wkbf[5]),olen));
    return(strlen(obuf));
}

void nkfVersion(){
    FILE *fp;

    sprintf(pop_cmd,"%s --version",config.nkf);
    if ( (fp=popen(pop_cmd,"r")) ==NULL) {
      return;
    }
    while(fgets(wkbf, sizeof(wkbf), fp) != NULL) {
      printf("%s",wkbf);
    }
    pclose(fp);
}

int isLastPacket(char * r_msg,int rLen){
    char wk[4];

    memcpy(wk,&r_msg[rLen-3],3);
    if(wk[0] == 0x2e && wk[1] == 0x0d && wk[2] == 0x0a) return(1);
    return(0);
}

int _send(int dmy, char *mbuf, int msize, int flags){
   if(ssl == NULL) return send(net, mbuf, msize, 0);
   else return SSL_write(ssl, mbuf, msize);
}

int _recv(int dmy, char *mbuf, int msize, int flags){
    if(ssl == NULL) return recv(net, mbuf, msize, 0);
    else return SSL_read(ssl, mbuf, msize);
}

void net_pop(char * buf){
    FILE *fp;
    char fname[64];

    sprintf(fname,"%s/.biff.log",config.cdir);
    if ((fp = fopen(fname,"r")) == NULL) {
      net_error("net_pop open()");
    }
    fscanf(fp,"%s",buf);
    fclose(fp);
}

void net_push(char * buf){
    FILE *fp;
    char fname[64];

    sprintf(fname,"%s/.biff.log",config.cdir);
    if ((fp = fopen(fname,"w")) == NULL) {
      net_error("net_push open()");
    }
    fprintf(fp,"%s",buf);
    fclose(fp);
}
int net_decode(char * fnamei, char * obuf, int olen){
    FILE *fp;
    char fnameo[21];
    char wk[128];

    strcpy(fnameo,"/tmp/bifftmp2XXXXXX");
    mktemp(fnameo);
if(DEBUG)printf("fnameo=[%s]\n",fnameo);
    sprintf(wk,"%s -w -O %s %s",config.nkf,fnamei,fnameo);
if(DEBUG)printf("wk=[%s]\n",wk);
    system(wk);

    if (( fp=fopen(fnameo,"r")) == NULL) {
      sys_error("net_decode:fopen-2()");
      return(0);
    }
    if (fgets(obuf, olen, fp) == NULL) {
      unlink (fnameo);
      sys_error("net_decode:fgets()");
      return(0);
    }
    fclose(fp);
    unlink (fnameo);
if(DEBUG)printf("obuf=[%s] strlen=%d\n",obuf,strlen(obuf));
    return(strlen(obuf));
}

int net_search(char * target, char * ptr, int ilen, char * obuf,int olen){
    char ch[1];
    char *ptrwk;
    int tlen=0;
    int fd;
    char fnamei[21];
    int	i,j;

if(DEBUG)printf("net_search:ilen=%d\n",ilen);
    for(i=0;i<ilen;i++) {
      memcpy(ch,ptr++,1);
      if(ch[0] != 0x0a) continue;
      if(strncmp(ptr,target,strlen(target)) == 0) break;
    }
    if (i == ilen) return(0); 
    ptrwk=ptr;
if(DEBUG)printf("net_search:i=%d,ptrwk=[%.10s]\n",i,ptrwk);

    while(1){
      memcpy(ch,ptr++,1);
      tlen++;
      if(ch[0] != 0x0a) continue;
      memcpy(ch,ptr++,1);
      if(ch[0] != 0x20) break;
      tlen++;
    }

if(DEBUG)printf("net_search:tlen=%d\n",tlen);
    strcpy(fnamei,"/tmp/bifftmp1XXXXXX");
    mktemp(fnamei);
    fd=open(fnamei,O_CREAT|O_RDWR|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH);
    if (fd == -1) {
      sys_error("open()");
      return(0);
    }
    for(i=0;i<tlen;i++) {
      memcpy(ch,ptrwk++,1);
      if(ch[0] == 0x0d) continue;
      if(ch[0] == 0x0a) continue;
      write(fd,ch,1);
    }
    close(fd);
    net_decode(fnamei,obuf,olen);
if(DEBUG)printf("net_search:obuf=[%s]\n",obuf);
    unlink (fnamei);

    return(strlen(obuf));
}

long net_list(int mes_no)
{
    char *p;
    int i=0;
    long lsize;

    sprintf(pop_cmd,"LIST %d\n",mes_no);
if(DEBUG)printf("pop_cmd=%s\n",pop_cmd);
    nbyte = strlen(pop_cmd);
    if( _send(net,pop_cmd,nbyte,0) < 0 ) net_error("send()");

//  change 2015/05/20
    while (1) {
      memset(r_msg,0,BUFMAX);
      if( _recv(net,r_msg,BUFMAX,0) < 0) net_error("recv()");
if(DEBUG)printf("LIST:[%s]",r_msg);
      if( strncmp(r_msg,"+OK",3) == 0) break;
    }

    p = strtok(r_msg, ",. ");
    while (p) {
if(DEBUG)printf("net_list:p=[%s] i=%d\n", p,i);
        if (i == 2) lsize=atol(p);
        p = strtok(NULL, ",. ");
        i++;
    }
if(DEBUG)printf("net_list:lsize=%d\n",lsize);
    return(lsize);
}

void net_dele(int mes_no)
{
    sprintf(pop_cmd,"DELE %d\n",mes_no);
if(DEBUG)printf("pop_cmd=%s\n",pop_cmd);
    nbyte = strlen(pop_cmd);
    if( _send(net,pop_cmd,nbyte,0) < 0 ) net_error("send()");

    memset(r_msg,0,BUFMAX);
    if( _recv(net,r_msg,BUFMAX,0) < 0) net_error("recv()");
if(DEBUG)printf("DELE:%s",r_msg);
}

int net_retr(int mes_no,long lsize,int viewMode)
{
    int i,j,k,l;
    int retCode=0;
    char *ptrOrg=NULL;
    char *ptrCur=NULL;
    char *tmp=NULL;
    int msize=1;
    int rsize=0;
    int rLen;
    char *ret1;
    char *ret2;

    if ((ptrOrg = (char *)malloc(msize)) == NULL) {
      sys_error("malloc()");
      return (0);
    }
    memset(ptrOrg,0,msize);

    sprintf(pop_cmd,"RETR %d\n",mes_no);
if(DEBUG)printf("net_retr:pop_cmd=%s\n",pop_cmd);
    nbyte = strlen(pop_cmd);
    if( _send(net,pop_cmd,nbyte,0) < 0 ) net_error("send()");

    while(1){
      memset(r_msg,0,BUFMAX);
      rLen = _recv(net,r_msg,BUFMAX,0);
      if( rLen < 0 ) net_error("recv()");
#if 0
if(DEBUG)printf("sizeof=%d rLen=%d\n",sizeof(r_msg),rLen);
if(DEBUG){
  for(i=MAX(rLen-10,0);i<rLen;i++){
    printf("%02x ",r_msg[i]);
  }
  printf("\n");
}
#endif
      msize=msize+BUFMAX;
      if ((tmp = (char *)realloc(ptrOrg,msize)) == NULL) {
        printf("realloc時にメモリが確保できません\n");
        free(ptrOrg);  /* 元のptrを解放して終了 */
        return (0);
      } else {
        ptrOrg = tmp;
//if(DEBUG)printf("net_retr:ptrOrg=%p\n",ptrOrg);
      /* reallocの戻り値は一度別変数に取り、
         NULLでないことを確認してから元の変数に代入するのが定石 */
      }
      ptrCur=ptrOrg+rsize;
//if(DEBUG)printf("net_retr:ptrOrg2=%p ptrCur=%p msize=%d rsize=%d\n",ptrOrg,ptrCur,msize,rsize);
//      memcpy(ptrCur,r_msg,sizeof(r_msg));
//      rsize=rsize+sizeof(r_msg);
      memcpy(ptrCur,r_msg,rLen);
      rsize=rsize+rLen;
//if(DEBUG)printf("net_retr:rsize=%d lsize=%d\n",rsize,lsize);
      ptrCur=ptrOrg+rsize;
      memset(ptrCur,0,1);
//      if (isLastPacket(r_msg,rLen)) break;
      if (rsize < lsize) continue;
      if (isLastPacket(ptrOrg,rsize)) break;
    }

//if(DEBUG)printf("net_retr:ptrOrg=[%s]\n",ptrOrg);
    for (i=0;i<config.mailn;i++) {
      memcpy(&mailwk,&config.mailp[i],sizeof(mailwk));
if(DEBUG)printf("mailwk.title[%d]=%s rsize=%d\n",i,mailwk.title,rsize);
      memset(buf,0,sizeof(buf));
      rLen=net_search("Subject:",ptrOrg,rsize,buf,sizeof(buf));
if(DEBUG)printf("net_retr:buf=[%s] rLen=%d\n",buf,rLen);
      if(viewMode) printf("[%s]\n",buf);
      if(strlen(mailwk.title)) {
        ret1=strstr(buf,mailwk.title);
      } else {
        ret1=(char *)1;
      }
if(DEBUG)printf("net_retr:mailwk.title=[%s] ret1=%d\n",mailwk.title,ret1);

if(DEBUG)printf("mailwk.from[%d]=%s\n",i,mailwk.from);
      memset(buf,0,sizeof(buf));
      rLen=net_search("From:",ptrOrg,rsize,buf,sizeof(buf));
if(DEBUG)printf("net_retr:buf=[%s] rLen=%d\n",buf,rLen);
      if(viewMode) printf("[%s]\n",buf);
      if(strlen(mailwk.from)) {
        ret2=strstr(buf,mailwk.from);
      } else {
        ret2=(char *)1;
      }
if(DEBUG)printf("net_retr:mailwk.from=[%s] ret2=%d\n",mailwk.from,ret2);
      if (ret1 != NULL && ret2 != NULL) {
        free (ptrOrg);
if(DEBUG)printf("net_retr:mailwk.del=%d mailwk.exit=%d\n",mailwk.del,mailwk.exit);
        if (mailwk.del == 1) net_dele(mes_no);
        net_push(buf);
        return (mailwk.exit);
      }
    }
    free (ptrOrg);
    return(0);
}

void net_quit(void) 
{
    sprintf(pop_cmd,"QUIT\n");
    nbyte = strlen( pop_cmd );
    if( _send(net,pop_cmd,nbyte,0) < 0) net_error("send()");

    memset(r_msg,0,BUFMAX);                    
    if( _recv(net,r_msg,BUFMAX,0) < 0) net_error("recv()");
if(DEBUG)printf("%s\n",r_msg);
}

void net_shut(ext_code) 
{
    if (ssl != NULL) {
      SSL_shutdown(ssl);
      SSL_free(ssl);		//SSL_new()で確保した領域解放
      SSL_CTX_free(ctx);	//SSL_CTX_new()で確保した領域解放
      ERR_free_strings();	//SSL_load_error_strings()で確保した領域解放
    }

    shutdown(net, 2);
    close(net);
    exit(ext_code);
}

int main(int argc, char *argv[])
{
    struct hostent *host;
    struct sockaddr_in sa;
    char work[10];
    int mes_no, mes_cnt=0;
    int loop;
    int wret,ext_code=0;
    int i;
    int rLen;
    char configFile[64];
    char curDir[64];
    int viewMode=0;
    int testMode=0;
    long lsize;


//  Get Config
if(DEBUG)printf("argv[0]=%s\n",argv[0]);
    strcpy(curDir,(char *)dirname(argv[0]));
if(DEBUG)printf("curDir=%s\n",curDir);
    sprintf(configFile,"%s/config.ini",curDir);
    if (argc >= 2) strcpy(configFile,argv[1]);
if(DEBUG)printf("configFile=%s\n",configFile);
    memset(&config,0,sizeof(config));
    if(ini_parse(curDir,configFile) != 0) {
      printf("config file[%s] not found\n",configFile);
      exit(0);
    }
    nkfPath(config.nkf,sizeof(config.nkf));
    if (argc == 3 && strcmp (argv[2],"TEST") == 0) {
      printf("config.port=%d\n",config.port);
      printf("config.cdir=%s\n",config.cdir);
      printf("config.host=%s\n",config.host);
      printf("config.user=%s\n",config.user);
      printf("config.pass=%s\n",config.pass);
      for(i=0;i<config.mailn;i++) {
        memcpy(&mailwk,&config.mailp[i],sizeof(mailwk));
        printf("config.from=%s\n",mailwk.from);
        printf("config.title=%s\n",mailwk.title);
        printf("config.del=%d\n",mailwk.del);
        printf("config.exit=%d\n",mailwk.exit);
      }
      testMode=1;
    }
    if (argc == 3 && strcmp (argv[2],"VIEW") == 0) viewMode=1;
if(DEBUG) printf("viewMode=%d\n",viewMode);

    if (testMode) {
      if (strlen(config.nkf)) {
        printf("nkf....OK\n");
        nkfVersion();
      } else {
        printf("nkf....Not install\n");
        exit(0);
      }
    }

//  Check host name
    if (testMode) printf("gethostbyname....");
    if ((host = gethostbyname(config.host)) == NULL){
      if (testMode) printf("Error\n");
      herror("gethostbyname()");
      exit(0);        
    }
    if (testMode) printf("OK\n");

    memset(&sa,0, sizeof(struct sockaddr_in));
    memcpy(&sa.sin_addr, host->h_addr, sizeof(struct in_addr));
    sa.sin_family = AF_INET;
//    sa.sin_port = htons(PORT_NO);
    sa.sin_port = htons(config.port);

//  Make socket
    if (testMode) printf("socket....");
    if((net = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
      if (testMode) printf("Error\n");
      net_error("socket()");
    }
    if (testMode) printf("OK\n");

//  Connect to the pop server
    if (testMode) printf("connect....");
    if (connect(net, (struct sockaddr *)&sa, sizeof(struct sockaddr)) < 0) {
      if (testMode) printf("Error\n");
      net_error("connect()");
    }


//for SSL
    if (config.port == POPS) {
      SSL_load_error_strings();
      SSL_library_init();
//      if (testMode) printf("SSL_CTX_new....");

//      if((ctx = SSL_CTX_new(TLSv1_client_method())) == NULL) {         //プロトコル選択。GmailはTLS
      if((ctx = SSL_CTX_new(SSLv23_method())) == NULL) {        /* make SSL_CONTEXT(SSLv2|SSLv3|TSLv1) */
        if (testMode) printf("Error\n");
        net_error("SSL_CTX_new()");
      }
//      if (testMode) printf("OK\n");

      SSL_CTX_set_options(ctx, SSL_OP_NO_SSLv2);  /* without SSLv2 */
//      if (testMode) printf("SSL_new....");
      if((ssl = SSL_new(ctx)) == NULL) {
        if (testMode) printf("Error\n");
        net_error("SSL_CTX_new()");
      }
//      if (testMode) printf("OK\n");

//      if (testMode) printf("SSL_set_fd....");
      if(SSL_set_fd(ssl, net) == 0) {
        if (testMode) printf("Error\n");
        net_error("SSL_set_fd()");
      }
//      if (testMode) printf("OK\n");

//      if (testMode) printf("SSL_connect....");
      if(SSL_connect(ssl) != 1) {
        if (testMode) printf("Error\n");
        net_error("SSL_connect()");
      }
//      if (testMode) printf("OK\n");
    }

    memset(r_msg,0,BUFMAX);
    if( _recv(net,r_msg,BUFMAX,0) < 0) net_error("recv()");
if(DEBUG)printf("CONNECT:%s",r_msg);
    if( strncmp(r_msg,"+OK",3) != 0) {
      if (testMode) printf("Error\n");
      net_shut(ext_code);
    }
    if (testMode) printf("OK\n");

//  Send user name
    if (testMode) printf("USER....");
    sprintf(pop_cmd,"USER %s\n",config.user);
    nbyte = strlen( pop_cmd );
    if( _send(net,pop_cmd,nbyte,0) < 0) net_error("send()");

    memset(r_msg,0,BUFMAX);
    if( _recv(net,r_msg,BUFMAX,0) < 0) net_error("recv()");
if(DEBUG)printf("USER:[%s]\n",r_msg);
    if( strncmp(r_msg,"+OK",3) != 0) {
      printf("Error\n");
      net_quit();
      net_shut(ext_code);
    }
    if (testMode) printf("OK\n");

//  Send password
    if (testMode) printf("PASS....");
    sprintf(pop_cmd,"PASS %s\n",config.pass);
    nbyte = strlen(pop_cmd);
    if( _send(net,pop_cmd,nbyte,0) < 0 ) net_error("send()");

    memset(r_msg,0,BUFMAX);
    if( _recv(net,r_msg,BUFMAX,0) < 0) net_error("recv()");
if(DEBUG)printf("PASS:[%s]\n",r_msg);
    if( strncmp(r_msg,"+OK",3) != 0) {
      printf("Error\n");
      net_quit();
      net_shut(ext_code);
    }
    if (testMode) printf("OK\n");

//  Request mail list
    if (testMode) printf("LIST....");
    sprintf(pop_cmd,"LIST\n");
    nbyte = strlen( pop_cmd );
    if( _send(net,pop_cmd,nbyte,0) < 0 ) net_error("send()");

    while(1){
      memset(r_msg,0,BUFMAX);
      rLen = _recv(net,r_msg,BUFMAX,0);
      if( rLen < 0 ) net_error("recv()");
//      if( _recv(net,r_msg,BUFMAX,0) < 0) net_error("recv()");
//      rLen=strlen(r_msg);
if(DEBUG)printf("LIST:rLen=%d [%s]\n",rLen,r_msg);
      if ( r_msg[0] == '.') break;

      if ( strncmp( r_msg, "+OK", 3) == 0 ) {
        memset(work,0,sizeof(work));
        for (loop=0 ; loop<sizeof(work) ; loop++) {
          if( r_msg[loop+4] == 0x20) break;
	  work[loop]=r_msg[loop+4];
	}
	mes_cnt = atoi(work);
if(DEBUG)printf("work=<%s> mes_cnt=%d\n",work,mes_cnt);
      }
      if (isLastPacket(r_msg,rLen)) break;
    }

    if (testMode) {
      printf("OK(number of mail=%d)\nAll test is OK!!\n",mes_cnt);
      net_quit();
      net_shut(ext_code);
    }

    if( mes_cnt == 0) {
      net_quit();
      net_shut(ext_code);
    }

//  Select letter number
    for (mes_no=1;mes_no<mes_cnt+1;mes_no++){  /* loop start */ 
      lsize=net_list(mes_no);
if(DEBUG)printf("mes_no=%d lsize=%d\n",mes_no,lsize);
      if (lsize == 0) continue;
      ext_code=net_retr(mes_no,lsize,viewMode);
if(DEBUG)printf("mes_no=%d ext_code=%d\n",mes_no,ext_code);
      if (ext_code != 0) break;
    } /* loop end */

    net_quit();
    net_shut(ext_code);
}
